from cobaya.likelihoods.base_classes import Planck2018Clik


class TE(Planck2018Clik):
    r"""
    High-$\ell$ cross TE power spectrum \textsc{plik} likelihood of Planck's 2018 data
    release \cite{Aghanim:2019ame}.
    """
    pass
