from cobaya.likelihoods.base_classes import Planck2015Clik


class planck_2015_lowTEB(Planck2015Clik):
    r"""
    Low-$\ell$ polarization likelihood of Planck's 2015 data release
    \cite{Aghanim:2015xee}.
    """
    pass
