from cobaya.likelihoods.base_classes import PlanckClik


class planck_2015_lensing(PlanckClik):
    r"""
    Lensing likelihood of Planck's 2015 data release based on temperature+polarization
    map-based lensing reconstruction \cite{Ade:2015zua}.
    """
    pass
