from cobaya.likelihoods.base_classes import PlanckPlikLite


class TTTEEE_lite_native(PlanckPlikLite):
    r"""
    High-$\ell$ temperature and polarization, foreground-marginalized likelihood of
    Planck's 2018 data release \cite{Aghanim:2019ame} (Python re-implementation of the
    official \textsc{plik\_lite}, by E.~Calabrese, and A.~Lewis).
    """
    pass
