from cobaya.likelihoods.base_classes import BAO


class sdss_dr16_baoplus_lrg(BAO):
    r"""
    Likelihood of the power spectrum LRG BAO of SDSS DR16 \cite{Alam:2020sor}.
    As recommended in the official data release, it combines measurements
    of LRGs from eBOSS DR16 and BOSS DR12.
    """
    pass
