from cobaya.likelihoods.base_classes import SN


class pantheon(SN):
    r"""
    Likelihood of the Pantheon type Ia supernova sample \cite{Scolnic:2017caz},
    including data from the Pan-STARRS1 (PS1) Medium Deep Survey.
    """
    pass
