from cobaya.likelihoods.base_classes import Planck2018Clik


class TT_unbinned(Planck2018Clik):
    r"""
    Unbinned high-$\ell$ temperature-only \textsc{plik} likelihood of Planck's 2018 data
    release \cite{Aghanim:2019ame}.
    """
    pass
