from cobaya.likelihoods.base_classes import BAO


class sdss_dr12_consensus_final(BAO):
    r"""
    Likelihood of the combined power spectrum BAO and full-shape measurements of
    SDSS DR12 \cite{Alam:2016hwk}.
    """
    pass
