from cobaya.likelihoods.base_classes import Planck2018Clik


class TTTEEE(Planck2018Clik):
    r"""
    High-$\ell$ temperature and polarization \textsc{plik} likelihood of Planck's 2018
    data release \cite{Aghanim:2019ame}.
    """
    pass
