from cobaya.likelihoods.base_classes import Planck2018CamSpecPython


class TTTEEE(Planck2018CamSpecPython):
    r"""
    High-$\ell$ temperature and polarization \textsc{CamSpec} likelihood of Planck's 2018
    data release \cite{Aghanim:2019ame} (native Python re-implemetation, by A.~Lewis).
    """
    pass
