from cobaya.likelihoods.base_classes import BAO


class sdss_dr7_mgs(BAO):
    r"""
    Likelihood of the BAO scale measurement of SDSS DR7 Main Galaxy Sample
    \cite{Ross:2014qpa}.
    """
    pass
