from cobaya.likelihoods.base_classes import Planck2018Clik


class TT_clik(Planck2018Clik):
    r"""
    Low-$\ell$ temperature-only likelihood of Planck's 2018 data release
    \cite{Aghanim:2019ame}.
    """
    pass
