from cobaya.likelihoods.base_classes import SN


class jla(SN):
    r"""
    Likelihood of the JLA type Ia supernova sample \cite{Betoule:2014frx}, based on
    observations obtained by the SDSS-II and SNLS collaborations.
    """
    pass
