from cobaya.likelihoods.base_classes import BAO


class sdss_dr16_lrg_bao_dmdh(BAO):
    r"""
    Likelihood of the power spectrum LRG BAO of SDSS DR16 \cite{Alam:2020sor}.
    It does not include the DR12 measurements.
    """
    pass
