from cobaya.likelihoods.base_classes import Planck2018Clik


class TTTEEE_unbinned(Planck2018Clik):
    r"""
    Unbinned high-$\ell$ temperature and polarization \textsc{plik} likelihood of Planck's
    2018 data release \cite{Aghanim:2019ame}.
    """
    pass
