from cobaya.likelihoods.planck_2018_highl_CamSpec2021.planck_2018_CamSpec2021_python import Planck2018CamSpec2021Python


class TTTE(Planck2018CamSpec2021Python):
    r"""
    High-$\ell$ TT + TE \textsc{CamSpec} 12.6HMcl likelihood using Planck 2018 data
    \cite{efstathiou2019}.
    """
    pass
