from cobaya.likelihoods.base_classes import BAO


class sdss_dr16_baoplus_lyauto(BAO):
    r"""
    Likelihood of the power spectrum Lyman alpha BAO of SDSS DR16 \cite{Alam:2020sor}.
    """
    pass
