from cobaya.likelihoods.planck_NPIPE_highl_CamSpec.planck_NPIPE_CamSpec_python import Planck2020CamSpecPython


class TE(Planck2020CamSpecPython):
    r"""
    High-$\ell$ TE \textsc{CamSpec} PR4-12.6cl likelihood using \texttt{NPIPE} (Planck PR4) data
    \cite{rosenberg22}.
    """
    pass
