from cobaya.likelihoods.base_classes import Planck2018Clik


class TTTEEE_lite(Planck2018Clik):
    r"""
    High-$\ell$ temperature and polarization, foreground-marginalized \textsc{plik\_lite}
    likelihood of Planck's 2018 data release \cite{Aghanim:2019ame}.
    """
    pass
