from cobaya.likelihoods.base_classes import PlanckPlikLite


class TT_lite_native(PlanckPlikLite):
    r"""
    High-$\ell$ temperature-only, foreground-marginalized likelihood of Planck's 2018 data
    release \cite{Aghanim:2019ame} (Python re-implementation of the official
    \textsc{plik\_lite}, by E.~Calabrese, and A.~Lewis).
    """
    pass
